/******************************************************************************
Asservissement de position d'un moteur à courant continu
Ce programme est destiné à fonctionner avec l'ensemble de 
commande de moteur à courant continu Raspberry C / Python, disponible à l'adresse suivante:
http://boutique.3sigma.fr/12-robots

Auteur: 3Sigma
Version 1.0 - 21/11/2016
*******************************************************************************/

// Inclusion d'une bibliothèque permettant l'exécution à cadence fixe
// d'une partie du programme. Télécharger à l'adresse http://www.3sigma.fr/telechargements/FlexiTimer2.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
// Pour plus de détails, voir les pages (en anglais): 
// http://www.arduino.cc/playground/Main/FlexiTimer2
// https://github.com/wimleers/flexitimer2
#include <FlexiTimer2.h>

// Inclusion de la bibliothèque de gestion de la carte AStar32U4 de contrôle moteur
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/AStar32U4.zip
// Pour plus de détails, voir la page (en anglais): 
// https://pololu.github.io/a-star-32u4-arduino-library/
#include <AStar32U4.h>

AStar32U4Motors motor;

// Définitions et déclarations pour le codeur incrémental du moteur
#define codeurPinA 0
#define codeurPinB 1
volatile long ticksCodeur = 0;
int16_t codeurDeltaPos;

// ATTENTION: donner la bonne valeur de la tension d'alimentation sur la ligne ci-dessous. Ici, 9 V
double tensionAlim = 9;

// Ce programme reçoit par liaison série des données issues d'un programme qui est exécuté sur
// l'ordinateur connecté à la carte. Si le programme ne reçoit rien pendant un temps donné,
// défini ci-dessous, un traitement particulier sera réalisé.
// Timeout de réception des données en s
#define TIMEOUT 2
unsigned long timeLastReceived = 0;
int timedOut = 0;

// Ce programme envoie par liaison série des données à l'ordinateur connecté à la carte.
// Cadence d'envoi des données en ms
#define TSDATA 20
unsigned long tempsDernierEnvoi = 0;
unsigned long tempsCourant = 0;

// Certaines parties du programme sont exécutées à cadence fixe grâce à la bibliothèque FlexiTimer2.
// Cadence d'échantillonnage en ms
#define CADENCE_MS 20
volatile double dt = CADENCE_MS/1000.;
volatile double temps = -CADENCE_MS/1000.;

// Le moteur est asservi en vitesse grâce à un régulateur de type PID
// On déclare ci-dessous les variables et paramètres nécessaires à l'asservissement et au régulateur
double thetaref = 0.; // position de consigne
double theta = 0.; // position du moteur
double commande = 0.; // commande en tension calculée par le PID
double commande_avant_sat = 0.; // valeur de la commande avant la saturation (voir ci-dessous)
double umax = 6.; // valeur max de la tension de commande du moteur
double umin = -6.; // valeur min (ou max en négatif) de la tension de commande du moteur
double yprec = 0; // mesure de la vitesse au calcul précédent
double Tf = 0.02; // constante de temps de filtrage de l'action dérivée du PID
double P_x = 0.; // valeur de l'action proportionnelle
double I_x = 0.; // valeur de l'action intégrale
double D_x = 0.; // valeur de l'action dérivée
// Variables intermédiaires
double Ti = 0;
double Td = 0;
double Tt = 0;
double ad = 0;
double bd = 0;
double br = 0;


// Déclaration des variables pour la réception des données sur la ligne série
const int NOMBRE_DE_CHAMPS = 8; // nombre de champs de données reçues sur la liaison série
int champIndex = 0; // champ courant reçu
int valeursRecues[NOMBRE_DE_CHAMPS]; // tableau contenant les champs de données reçues
// Variables utilisées pour les données reçues
int typeSignal = 0;
double offset = 0.;
double amplitude = 0.;
double frequence = 0.;
double Kp = 1.0; // gain proportionnel du PID
double Ki = 0.1; // gain intégral du PID
double Kd = 0.05; // gain dérivé du PID
// Variables pour la vérification de cohérence des données
double crc = 0.;
double crc_check = 0.;


// Initialisations
void setup(void) {
  
  // Codeur incrémental moteur
  FastGPIO::Pin<codeurPinA>::setInputPulledUp();
  FastGPIO::Pin<codeurPinB>::setInputPulledUp();
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurPinA (définie à la fin du programme)
  attachInterrupt(digitalPinToInterrupt(codeurPinA), GestionInterruptionCodeurPinA, CHANGE);
  // A chaque changement de niveau de tension sur le pin B du codeur,
  // on exécute la fonction GestionInterruptionCodeurPinB (définie à la fin du programme)
  attachInterrupt(digitalPinToInterrupt(codeurPinB), GestionInterruptionCodeurPinB, CHANGE);

  // Liaison série.
  Serial.begin(115200);
  while (!Serial) {
    ; // Attente de la connexion du port série. Requis pour l'USB natif
  }
  
  /* Le programme principal est constitué:
     - de la traditionnelle boucle "loop" des programmes Arduino, dans laquelle on lit 
       en permanence la liaison série pour récupérer les nouvelles consignes de tension
       de commande du moteur
     - de la fonction "isrt", dont l'exécution est définie à cadence fixe par les deux instructions
       ci-dessous. La bonne méthode pour exécuter une fonction à cadence fixe, c'est de l'exécuter sur interruption
       (la boucle "loop" est alors interrompue régulièrement, à la cadence voulue, par un timer).
       Une mauvaise méthode serait d'utiliser la fonction Arduino "delay". En effet, la fonction "delay" définit
       un délai entre la fin des instructions qui la précèdent et le début des instructions qui la suivent, mais si
       on ne connait pas le temps d'exécution de ces instructions (ou si celui-ci est inconnu), on n'obtiendra
       jamais une exécution à cadence fixe.
       Il est nécessaire d'exécuter la fonction "isrt" à cadence fixe car cette fonction:
       - doit renvoyer des données à cadence fixe au programme exécuté sur l'ordinateur connecté à la carte
       - calcule la vitesse de rotation du moteur en comptant le nombre d'impulsions du codeur pendant un temps fixe
     
  */
  FlexiTimer2::set(CADENCE_MS, 1/1000., isrt); // résolution timer = 1 ms
  FlexiTimer2::start();

}

// Boucle principale
void loop() {

  // Réception des données sur la liaison série
  readData();
  
}

// Fonction excutée sur interruption
void isrt(){

  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurDeltaPos += ticksCodeur;
  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeur = 0;
  
  // Calcul de la position de rotation. C'est le nombre d'impulsions converti en radian
  theta = -2.*3.141592*((double)codeurDeltaPos)/1632.;  // en rad
  
  // Calcul de la consigne en fonction des données reçues sur la liaison série
  if (typeSignal==0) { // signal carré
    if (frequence > 0) {
      if (temps - ((double)((int)(temps*frequence)))/frequence < 1/(2*frequence)) {
        thetaref = offset + amplitude;
      }
      else {
        thetaref = offset;
      }
    }
    else {
      thetaref = offset + amplitude;
    }
  }
  else { // sinus
    if (frequence > 0) {
      thetaref = offset + amplitude * sin(2*3.141592*frequence*temps);
    }
    else {
      thetaref = offset + amplitude;
    }
  }
  
  /******* Calcul du PID *******/
  // Paramètres intermédiaires
  Ti = Ki/(Kp+0.01);
  if (Kd>0) { // Si PID
    ad = Tf/(Tf+dt);
    bd = Kd/(Tf+dt);
    Td = Kp/Kd;
    Tt = sqrt(Ti*Td);
  }
  else { // Si PI
    Tt = 0.5*Ti;
  }
  br = dt/(Tt+0.01);
  
  // Terme proportionnel
  P_x = Kp * (thetaref - theta);
  
  // Terme dérivé
  D_x = ad * D_x - bd * (theta - yprec);
  
  // Calcul de la commande avant saturation
  commande_avant_sat = P_x + I_x + D_x;
  
  // Application de la saturation sur la commande
  if (commande_avant_sat > umax) {
    commande = umax;
  }
  else if (commande_avant_sat < umin) {
    commande = umin;
  }
  else {
    commande = commande_avant_sat;
  }
  
  // Terme intégral (sera utilisé lors du pas d'échantillonnage suivant)
  I_x = I_x + Ki * dt * (thetaref - theta) + br * (commande - commande_avant_sat);
  
  // Stockage de la mesure courante pour utilisation lors du pas d'échantillonnage suivant
  yprec = theta;

  /******* Fin Calcul du PID *******/
      
  // Envoi de la commande au moteur
  // Si le timeout de réception des données a été déclenché (voir plus loin),
  // on met la commande à 0
  if (timedOut==1) {
    commande = 0.;
  }
        
  // Envoi de la commande au moteur
  CommandeMoteur(commande, tensionAlim);

  // Incrémentation du temps courant
  temps += dt;
}

void ecritureData(void) {

  // Ecriture des données en sortie tous les TSDATA millisecondes
  
  // On envoie les données s'il s'est écoulé plus de TSDATA secondes
  // entre l'instant courant et le dernier envoi
  tempsCourant = millis();
  if (tempsCourant-tempsDernierEnvoi > TSDATA) {
    Serial.print(temps);
    
    Serial.print(",");
    Serial.print(thetaref);
    Serial.print(",");
    Serial.print(theta);
    Serial.print(",");
    Serial.print(commande);

   
    Serial.print("\r");
    Serial.print("\n");
    
    tempsDernierEnvoi = tempsCourant;
  }  
}

void readData(void) {
  // Lecture des données sur la liaison série
  // On attend une série de NOMBRE_DE_CHAMPS valeurs entières séparées par des virgules
  
  // Initialisations
  champIndex = 0;
  for(int i=0; i < NOMBRE_DE_CHAMPS; i++) {
    valeursRecues[i] = 0;
  }

  while(true) {

    // Ecriture des données sur la liaison série
    ecritureData();
  
    if (Serial.available()>0) { // si une donnée est reçue
      // Réinitialisation du timeout (voir plus loin)
      timeLastReceived = millis();
      timedOut = 0;
      // Lecture de la donnée
      char ch = Serial.read();
      if (ch >= '0' && ch <= '9') { // caractère ascii entre 0 et 9 ?
        // Si oui, nous stockons la valeur
        valeursRecues[champIndex] = (valeursRecues[champIndex] * 10) + (ch - '0');
      }
      else if (ch == ',') { // séparateur virgule détecté, on passe au champ suivant
        if(champIndex < NOMBRE_DE_CHAMPS-1) {
          champIndex++; // incrémentation de l'index de champ
        }
      }
      else {
        // tout caractère autre qu'un chiffre ou une virgule (le retour chariot de fin de trame par exemple)
        // stoppe l'acquisition des champs
        break; // on sort de la boucle
      } 
    }
    // Timeout
    // Si aucune donnée n'a été reçue pendant plus de TIMEOUT ms, on met à zéro toutes les
    // consignes par sécurité
    else { 
      if ((millis()-timeLastReceived > TIMEOUT*1000) && (timedOut==0)) {
        for(int i=0; i < NOMBRE_DE_CHAMPS; i++) {
          offset = 0.;
          amplitude = 0.;
          frequence = 0.;
        }
        timedOut = 1;
      }
    }    
  }
  
  // Traitement des données reçues (on extrait chaque valeur de la trame complète)
  if (champIndex == NOMBRE_DE_CHAMPS-1) { // si le bon nombre de champs a été reçu
      
    // Récupération de la dernière valeur (somme de contrôle, envoyée par l'émetteur)
    crc = valeursRecues[NOMBRE_DE_CHAMPS-1];
    // Calcul de la somme de contrôle (somme de toutes les autres données) côté récepteur
    crc_check = 0;
    for(int i=0; i < NOMBRE_DE_CHAMPS-1; i++) {
      crc_check += valeursRecues[i];
    }
    
    // Si les sommes de contrôle sont identiques, on récupère les données
    if (crc == crc_check) {
      typeSignal = valeursRecues[0];
      // Conversion des données d'entiers en flottants
      offset = ((double)valeursRecues[1]-100)/10.;
      amplitude = ((double)valeursRecues[2])/10.;
      frequence = ((double)valeursRecues[3])/100.;
      Kp = ((double)valeursRecues[4])/100.;
      Ki = ((double)valeursRecues[5])/100.;
      Kd = ((double)valeursRecues[6])/1000.;
    }
    // Si les sommes de contrôle sont différentes, on fait un traitement particulier
    else {
      // Par exemple, envoi d'un message, désactivé ici
      //Serial.print("\r\r\rErreur CRC !!!");
    }
  }
  else { // on n'a pas reçu le bon nombre de champs
    // On fait un traitement particuliern pas exemple, envoi d'un message (désactivé ici)
    //Serial.print("\r\r\rErreur reception !!!\r\r\r");
  }
}

void CommandeMoteur(double commande, double tensionAlim) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;
  
  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commande(commande);
  
  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(400*(tension/tensionAlim));
  
  // Saturation par sécurité
  if (tension_int>400) {
    tension_int = 400;
  }
  if (tension_int<-400) {
    tension_int = -400;
  }
  
  // Commande PWM
  motor.setM1Speed(tension_int);
}

double redresse_commande(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.); 
  }
  else if (commande<0) {
    return(commande-0.); 
  }
  else {
    return(0); 
  }
}
 
void GestionInterruptionCodeurPinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental
  // On utilise la fonction isInputHigh de la librairie FastGPIO
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (FastGPIO::Pin<codeurPinA>::isInputHigh() == FastGPIO::Pin<codeurPinB>::isInputHigh()) {
    ticksCodeur++;
  }
  else {
    ticksCodeur--;
  }
}


void GestionInterruptionCodeurPinB() {  
  // Routine de service d'interruption attachée à la voie B du codeur incrémental
  // On utilise la fonction isInputHigh de la librairie FastGPIO
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (FastGPIO::Pin<codeurPinA>::isInputHigh() == FastGPIO::Pin<codeurPinB>::isInputHigh()) {
    ticksCodeur--;
  }
  else {
    ticksCodeur++;
  }
}

